#include <18f4520.h>
#device high_ints=TRUE                                                  //Enable high priority interrupts
#fuses HS,NOLVP,NOWDT,NOPROTECT
#use delay(clock=40000000)                                              //Numbers are preset for 40 mHz clock
#use i2c(SLAVE, FAST, SCL=PIN_C3, SDA=PIN_C4, address=0x80, FORCE_HW)   //Using hardware I2C, defaults to address 0x80 or device 1 


//I2C variables
int State;                                            //State of Transmission
int32 Number32Bit = 0;                                //Stores bumbers converted to 32 bit
int Number[4] = {0,0,0,0};                            //Stores 4 bytes of 32 bit numbers 
int DataType = 0;                                     //Type of operation being preformed


//Encoder Variables
int32 TotalCount = 1<<31;
int16 Count0 = 0, Count1 = 0;
int16 Last0 = 0, Last1 = 0;

//PID vars
signed int32 error = 0;
signed int32 p_error = 0;
int32 kp = 1;
int32 ki = 0;
int32 kd = 0;
signed int32 Duty = 511;

//Velocity Variables
signed int32 Velocity = 0;

//Tracking Variables
int TrackingMode = 0;
signed int32 TargetVelocity = 0;
int32 TargetPosition = (1<<31);

// Trapezoidal Tracking Variables
signed int32 vel_max = 0;
signed int32 acc_desired = 0;
signed int32 vel_desired = 0;
int32 pos_desired = 1<<31;
signed int32 pos_error = 0;

//signed int32 pos_desired = 0;
int32 actual_time = 0;
int32 t1 = 0;
int32 t2 = 0;
int32 t3 = 0;
int32 d1, d2, d3;

// Function Prototypes
void runsetup();
void trapezoidalTrack();
void PID();
void recombine();

// Set up PIC's internal hardware
void runsetup()
{ 
   I2C_SlaveAddr(128|((int)input(pin_B4)<<1)|((int)input(pin_B5)<<2)|((int)input(pin_E2)<<3)|((int)input(pin_E1)<<4));//setting internal address see data sheet for truth table
   
   setup_timer_0(RTCC_EXT_L_TO_H | RTCC_DIV_1);       //external input from quadrature encoder 0
   setup_timer_1(T1_EXTERNAL | T1_DIV_BY_1);          //external input from quadrature encoder 1
   setup_timer_2(T2_DIV_BY_1, 255, 1);                //for PWM at 20KHz, out of audible range
   setup_timer_3(T3_INTERNAL|T3_DIV_BY_1);            //for counting time, increments every .2us and rolls over 76.3 times per sec
   enable_interrupts(INT_SSP);                        //enable I2C interrupt
   enable_interrupts(INT_EXT);                        //enable "go" interrupt on pin C7
   enable_interrupts(global);
   
   setup_ccp2(ccp_pwm);                               //setup PWM
   set_timer0(0);                                     //clear starting point of timers
   set_timer1(0);
   set_PWM2_duty(511);                                //motor starts at 0 velocity
}

// Pre-process the trapezoidal profile parameters
void trapezoidalCompute()
{
   pos_error = (signed int32) (pos_desired - TotalCount);
   
   acc_desired = 1;
   vel_max = abs(vel_max);
   if (pos_error < 0)
   {
      vel_max = -vel_max;
      acc_desired = -1;
   }
      
   t1 = vel_max/ acc_desired;
   d1 = t1 * t1 * acc_desired;
   d1 /= 2;
   
   d2 = pos_error - d1 - d1;
   t2 = (d2 / vel_max) + t1;
   d2 += d1;
   
   t3 = t2 + t1;
   d3 = d1 + d2;
}

// Perform the trapezoidal increasing of position
// Using Euler approximation
void trapezoidalTrack()                               // Trapezoidal Position Tracking with vel/acc control
{
   if (actual_time < t1)                              // Phase 1, position increasing, velocity increasing
   {
      vel_desired += acc_desired;
      TargetPosition += vel_desired;
   }
   else if (actual_time < t2)                         // Phase 2, position increasing, velocity constant
   {
      vel_desired = vel_max;
      TargetPosition += vel_desired;
   }
   else if (actual_time < t3)                         // Phase 3, position increasing, velocity decreasing
   {
      vel_desired -= acc_desired;
      TargetPosition += vel_desired;
   }
   else                                               // Phase 4, position constant, velocity 0
   {
      TargetPosition = pos_desired;
   }
   
   // Increase the time variable
   actual_time++;
   // PID the new position
   PID();
}

//Proportional position control
void PID()                                            
{
   error = TargetPosition - TotalCount;         //Compute p error
   Duty = 511;
   p_error = error;
   Duty += p_error;                             //Adjust Duty of motor (wrt to error)
   
   if (Duty > 1023)                             //Check for Saturation
   {
      Duty = 1023;
   }
   else if (Duty < 0)
   {
      Duty = 0;
   }
 
   set_PWM2_duty((int16)Duty);                               //Adjust PWM in response to PID control
}

void cleartrackingvars()                              //clears all tracking variables on each new command
{
   Duty = 511;                                        //Motors stop
   error = 0;                                         //Error variables set to 0
   p_error = 0;
}

void updatestatus()                                   //Calculates position and velocity based on encoder counts
{
   Count0 = get_timer0();                             //Up counts
   Count1 = get_timer1();                             //Down counts
   
   //Velocity = ((signed int32)(Count0 - Last0))-((signed int32)(Count1 - Last1));    //Current Velocity
   
   disable_interrupts(global);                        //Disable interrupts temporarily to prevent inacurate position calcultaion
   TotalCount += Count0;
   TotalCount -= Last0;           //Increment up counts
   if (Count0 < Last0)
   {
      TotalCount += 65536;
   }
   Last0 = Count0;
   
   TotalCount -= Count1;
   TotalCount += Last1;           //Increment down counts
   if (Count1 < Last1)
   {
      TotalCount -= 65536;
   }  
   Last1 = Count1;       
   enable_interrupts(global);                         //Re-enable interrupts

   TrapezoidalTrack();                                //Track with Trapezoidal Trakcing
}

void updatetracking()                                 //Updates variables on external signal on line B0
{
   switch(DataType)
   {
      case 1:                                         //Position or Velocity select
         TrackingMode = (int8)Number32Bit;
      break;
      case 2:                                         //Target position update
         pos_desired = (1<<31)+(signed int32)Number32Bit;
         trapezoidalCompute();
         actual_time = 0;
      break;
      case 3:                                         //Target velocity update
         TargetVelocity = (signed int32)Number32Bit;
         vel_max = TargetVelocity;
      break;
      case 4:                                         //Proportionality constant update
         kp = (int16) Number32Bit;
      break;
      case 5:                                         //Integral constant update
         ki = (int16) Number32Bit;
      break;
      case 6:                                         //Derative constant update
         kd = (int16) Number32Bit;
      break;
      case 10:                                        //Reset position variable
         TotalCount = 1<<31;
         //TargetVelocity = 0;
         TargetPosition = 1<<31;
         
         vel_max = 0;
         acc_desired = 0;
         vel_desired = 0;
         pos_desired = 1<<31;
         t1 = 0;
         t2 = 0;
         t3 = 0;
         
      break;
   }
   cleartrackingvars();                               //Clear tracking variables after new data is received
}

void recombine()                                      //Turns 4 one byte numbers into one 32 bit number
{
   Number32Bit = ((int32)Number[0]|(int32)Number[1]<<8|(int32)Number[2]<<16|(int32)Number[3]<<24);
   delay_us(1);                                       //Prevents overflow
}

#INT_EXT high
void motorgo()                                        //Begins latest tracking information
{
   updatetracking();
}

#INT_SSP high
void ssp_interupt ()                                  //I2C interrupt
{
   disable_interrupts(global);                        //Disable interrupts necessary to prevent exit during transmission
   state = i2c_isr_state();
   if(state < 0x80)                                   //Master is sending data
   {
      if(state == 0)                                  //First received byte is address
      {
      }
      if(state == 1)                                  //Second received byte is address
      {
         DataType = i2c_read();
      }
      if(state >1)                                    //Additional received bytes are data
      {
         Number[state-2] = i2c_read();
      }
      if (state==5)                                   //Combine 4 one byte numbers into one 32 bit number
      {
      recombine();
      }
   }
   if(state >= 0x80)                                  //Master is requesting data
   {
      switch(DataType)                                //Slices data to send to master
      {
         case 255:
            i2c_write((int8)(TotalCount));
         break;
         case 254:
            i2c_write((int8)(TotalCount>>8));
         break;
         case 253:
            i2c_write((int8)(TotalCount>>16));
         break;
         case 252:
            i2c_write((int8)(TotalCount>>24));
         break;
         case 251:
            i2c_write((int8)(Velocity));
         break;
         case 250:
            i2c_write((int8)(Velocity>>8));
         break;
         case 249:
            i2c_write((int8)(Velocity>>16));
         break;
         case 248:
            i2c_write((int8)(Velocity>>24));
         break;
         case 247:
            i2c_write((int8)(TargetPosition));
         break;
         case 246:
            i2c_write((int8)(TargetPosition>>8));
         break;
         case 245:
            i2c_write((int8)(TargetPosition>>16));
         break;
         case 244:
            i2c_write((int8)(TargetPosition>>24));
         break;
         case 243:
            i2c_write((int8)(TargetVelocity));
         break;
         case 242:
            i2c_write((int8)(TargetVelocity>>8));
         break;
         case 241:
            i2c_write((int8)(TargetVelocity>>16));
         break;
         case 240:
            i2c_write((int8)(TargetVelocity>>24));
         break;
      }
   }
   enable_interrupts(global);                         //Re-enable interrupts
}

void main() 
{
   runsetup();
   
   while (TRUE) 
   {
      output_d(pos_error);
      
      while(get_timer3()>4950)                        //Enters main function every 500us or 2000 times per second
      {
         set_timer3(0);                               //Reset timer 3
         DISABLE_INTERRUPTS(INT_TIMER3);
         updatestatus();
         ENABLE_INTERRUPTS(INT_TIMER3);
      }
   }
}
